# ICFP 2002 entry for OaSys
# see http://icfpcontest.cse.ogi.edu/ for details of contest
# Contact: Terry Stewart at terry@oasys.ca
#
# Overview:
#  A fairly simple little program.  Here's the basic logic:
#     if we can drop packages to get points, do so
#     if we can pick up packages, pick up as many as possible (weighted by distance to destination)
#     if we have packages, go towards the closest one's destination
#     otherwise, go to the nearest base (ignoring ones we are pretty sure are empty)
#  One neat trick: When going towards a particular location, it will try to avoid running
#    into other robots (as long as that doesn't slow it down getting to its target)
#  Finding a path to a location is done by determining the distance from every point on the map to
#    the target (and caching this data, of course)
#

import random
import socket
import sys


# ------------------------------------
# COMMUNICATIONS functions
# ------------------------------------


# initialize socket
s=socket.socket(socket.AF_INET, socket.SOCK_STREAM)
s.connect((sys.argv[1],int(sys.argv[2])))

def sendLine(text):
  "Adds a \\n and sends the text to the server"
  print '--->'+text
  text=text+'\n'
  v=s.send(text)
  while v!=len(text):
    text=text[v:]
    v=s.send(text)
    

extra=''
def readLine():
  "Returns one line from the server"
  global extra
  while '\n' not in extra:
    extra=extra+s.recv(1024)
  i=extra.find('\n')
  res=extra[:i]
  extra=extra[i+1:]
  print '    S:',res
  return res

def sendMove(bid,dir):
  "requests a move in a given direction"
  if dir not in ('N','S','E','W'):
    print 'ERROR! tried to move',dir
  sendLine('%d Move %s' % (bid,dir))
  
def sendPick(bid,packs):
  "attempts to pick up packages"
  m='%d Pick' % bid
  for p in packs:
    m=m+' '+`p`
  sendLine(m)
  
def sendDrop(bid,packs):
  "attempts to drop packages"
  m='%d Drop' % bid
  for p in packs:
    m=m+' '+`p`
  sendLine(m)
    

# ------------------------------------
# Board data
# ------------------------------------

class Board:
  def __init__(self):
    self.g={}           # cached gradient maps for path finding
    self.packWeight={}  # known weights of packages
    self.packDest={}    # known destinations of packages
    self.bases=[]       # list of bases that may have packages
    self.score={}       # best guesses for all players' scores
    
    self.w=-1           # width of board (after adding walls all around)
    self.h=-1           # height of board (after adding walls all around)
    self.map=[]         # map of the world (with walls added all around).  Access with map[y][x].
    self.id=''          # id of the robot (of the form '#3')
    self.strength=-1    # how much the robot can carry
    self.money=-1       # the amount of money the robot has
    self.loc={}         # location and carry list of all known robots.  loc['#1']=[x,y,[p1,p2,p3]]       
    
  def calcGradient(self,sx,sy):
    """makes sure a gradient map exits for the given (x,y) location.
       Calculates this by setting the distance to the target of point (x,y) to 0
       and then making all those points around it a distance of 1 (other than walls
       and water), and so on."""
    if (sx,sy) in self.g.keys():
      return
    gr=[]
    for i in range(0,self.h):
      gr.append([99999]*self.w)
    l=[(sx,sy)]
    gr[sy][sx]=0
    while len(l)>0:
      (x,y)=l.pop(0)
      val=gr[y][x]
      if gr[y+1][x]==99999 and self.map[y+1][x] not in ('#','~'):
        l.append((x,y+1))
        gr[y+1][x]=val+1
      if gr[y-1][x]==99999 and self.map[y-1][x] not in ('#','~'):
        l.append((x,y-1))
        gr[y-1][x]=val+1
      if gr[y][x+1]==99999 and self.map[y][x+1] not in ('#','~'):
        l.append((x+1,y))
        gr[y][x+1]=val+1
      if gr[y][x-1]==99999 and self.map[y][x-1] not in ('#','~'):
        l.append((x-1,y))
        gr[y][x-1]=val+1
    self.g[(sx,sy)]=gr
    
  def getDistanceTo(self,x,y,exact=0):
    """returns the length of the best path from the current location to (x,y).
       If exact is zero, then it modifies this to add one to a distance if
       a robot is there or one step away."""
    (fx,fy,p)=self.loc[self.id]
    return self.getDistanceToFrom(x,y,fx,fy,exact)

  def getDistanceToFrom(self,x,y,fx,fy,exact=0):
    """returns the length of the best path from (fx,fy) to (x,y).
       If exact is zero, then it modifies this to add one to a distance if
       a robot is there or one step away."""
    self.calcGradient(x,y)
    r=self.g[(x,y)][fy][fx]
    if not exact:
      if self.isOtherAt(fx,fy):
        r=r+1
      elif self.isOtherNear(fx,fy):
        r=r+1
    return r

  def isOtherAt(self,xx,yy):
    "checks if there is a robot other than me at (xx,yy)"
    for id in self.loc.keys():
      (x,y,p)=self.loc[id]
      if (x,y)==(xx,yy) and id!=self.id:
        return 1
    return 0

  def isOtherNear(self,x,y):
    "checks if there is a robot other than me anywhere around (x,y)"
    return self.isOtherAt(x+1,y) or self.isOtherAt(x-1,y) or self.isOtherAt(x,y+1) or self.isOtherAt(x,y-1)
  
  def getCarriedWeight(self):
    "returns the amount of weight of packages the robot is currently carrying"
    total=0
    for i in self.loc[self.id][2]:
      total=total+self.packWeight[i]
    return total

  def findBases(self):
    "scans the map for bases"
    self.bases=[]
    for y in range(0,len(self.map)):
      x=-1
      while 1:
        x=self.map[y].find('@',x+1)
        if x==-1:
          break
        else:
          self.bases.append((x,y))
  
  
# ------------------------------------
# Game functions
# ------------------------------------
  
def readBoard():
  "reads the initialization data from the server and sets up the board"  
  b=Board()
  dim=readLine().split()
  (b.w,b.h)=(int(dim[0])+2,int(dim[1]))
  b.map=['#'*b.w]
  for i in range(0,b.h):
    b.map.append('#'+readLine()+'#')
  b.map.append('#'*b.w)
  b.h=b.h+2
  b.findBases()
  
  me=readLine().split()
  (b.id,b.strength,b.money)=('#'+me[0],int(me[1]),int(me[2]))
  
  b.loc={}
  loc=readLine().split()
  while len(loc)>0:
    (id,xn,x,yn,y)=loc[:5]
    loc=loc[5:]
    b.loc[id]=[int(x),int(y),[]]
  return b

def readPackageList(b):
  "reads the list of packages at the current location"
  p=readLine().split()
  l={}
  while len(p)>0:
    (id,x,y,w)=p[:4]
    p=p[4:]
    id=int(id)
    x=int(x)
    y=int(y)
    w=int(w)
    l[id]=(x,y,w)
    b.packWeight[id]=w
    b.packDest[id]=(x,y)
  return l

def readUpdate(b):
  "reads the data from the server about what happened last turn, and updates the board"
  com=readLine().split()
  ids=[]
  while len(com)>0:
    id=com.pop(0)
    ids.append(id)
    acts=[]
    while len(com)>0 and com[0][0]!='#':
      acts.append(com.pop(0))
    while len(acts)>0:
      act=acts.pop(0)
      if act=='N':
        b.loc[id][1]=b.loc[id][1]+1
      elif act=='S':
        b.loc[id][1]=b.loc[id][1]-1
      elif act=='E':
        b.loc[id][0]=b.loc[id][0]+1
      elif act=='W':
        b.loc[id][0]=b.loc[id][0]-1
      elif act=='P':
        p=int(acts.pop(0))
        b.loc[id][2].append(p)
      elif act=='D':
        p=int(acts.pop(0))
        if p in b.packDest.keys():
          print b.packDest[p],b.loc[id]
          if b.packDest[p]==(b.loc[id][0],b.loc[id][1]):
            if id not in b.score.keys(): b.score[id]=0
            b.score[id]=b.score[id]+b.packWeight[p]
        if p in b.loc[id][2]:
          b.loc[id][2].remove(p)
      elif act=='X':
        if id not in b.loc.keys():
          b.loc[id]=[-1,-1,[]]
        x=int(acts.pop(0))
        b.loc[id][0]=x
      elif act=='Y':
        if id not in b.loc.keys():
          b.loc[id]=[-1,-1,[]]
        y=int(acts.pop(0))
        b.loc[id][1]=y
  for i in b.loc.keys():
    if i not in ids:
      del b.loc[i]
      if i in b.score.keys():
        del b.score[i]        

def listPossibleDrop(b):
  "return all the held packages that have the current square as their location"
  (x,y,p)=b.loc[b.id]
  return filter(lambda id,b=b,x=x,y=y: b.packDest[id]==(x,y),p)
  
def getDest(b):
  "return the closest destination for a package I'm holding.  Returns None if there are no packages."
  (x,y,p)=b.loc[b.id]
  best=None
  dist=99999
  for id in p:
    (x,y)=b.packDest[id]
    d=b.getDistanceTo(x,y)
    if d<dist:
      best=(x,y)
      dist=d
  return best  
  
def listPossiblePackages(b,pack):
  """returns a list of packages we can pick up.  Tries to optimize this a bit
     by using a greedy algorithm and scales the weights by the distance needed
     to travel to deliver it and get back.  Not the best algorithm in the
     world, but it'll do for now."""
  max=b.strength-b.getCarriedWeight()
  grab=[]
  all=[]
  for id in pack.keys():
    (x,y)=b.packDest[id]
    w=b.packWeight[id]
    value=float(w)/(2*b.getDistanceTo(x,y))
    all.append((id,value,w))
  all.sort(lambda a,b:cmp(b[1],a[1]))
  for (id,v,w) in all:
    if w<=max:
      grab.append(id)
      max=max-w
  return grab
    

def goTo(b,x,y,bid=1):
  """moves towards (x,y).  Checks to see which direction will make it closest to
     the target.  Randomly chooses among equal distances.  Since getDistanceToFrom
     increases by one the distance measurement if there's another robot near that spot,
     it will tend to avoid other robots if there is another equally good route."""
  (sx,sy,p)=b.loc[b.id]
  n=b.getDistanceToFrom(x,y,sx,sy+1)
  s=b.getDistanceToFrom(x,y,sx,sy-1)
  e=b.getDistanceToFrom(x,y,sx+1,sy)
  w=b.getDistanceToFrom(x,y,sx-1,sy)
  best=min(n,s,e,w)
  l=filter(lambda (v,s),best=best: v==best,((n,'N'),(s,'S'),(e,'E'),(w,'W')))
  (value,symbol)=random.choice(l)
  sendMove(bid,symbol)


def doAI(b,pack):
  "Does the AI for a turn."
  
  # if I'm at a base, and there's nothing here, remove it from my list of bases.
  (sx,sy,p)=b.loc[b.id]
  if len(pack)==0 and b.map[sy][sx]=='@' and (sx,sy) in b.bases:
    print 'Removing base at (%d,%d) from list' % (sx,sy)
    b.bases.remove((sx,sy))

  print b.id,b.money,b.score    
  print b.loc
  print 'Robot at (%d,%d)' % (b.loc[b.id][0],b.loc[b.id][1])
  
  # try to drop packages if we're at the destination
  drop=listPossibleDrop(b)
  if len(drop)>0:
    print 'Dropping packages at target:',drop
    sendDrop(1,drop)
  else:

    # try to grab packages if we can
    grab=listPossiblePackages(b,pack)
    if len(grab)>0:
      print 'Picking up packages:',grab
      sendPick(1,grab)

    else:
      # go towards the closest destination for a package we have
      dest=getDest(b)
      if dest!=None:
        print 'Delivering package to (%d,%d)'%(dest[0],dest[1])
        goTo(b,dest[0],dest[1])

      else:
        # go towards the nearest base
        bbd=99999
        if len(b.bases)==0:
          b.findBases()
        for base in b.bases:
          d=b.getDistanceTo(base[0],base[1])
          if d<bbd:
            bbd=d
            bb=base
        print 'Going to base at (%d,%d)' % (bb[0],bb[1])
        goTo(b,bb[0],bb[1])   
  b.money=b.money-1    #we always bid 1
      
      
      
 

#play the game
sendLine('Player')
b=readBoard()
while 1:
  pack=readPackageList(b)
  doAI(b,pack)
  readUpdate(b)


